'' =================================================================================================
''
''   File....... jm_i2c.spin
''   Purpose.... Basic I2C driver
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2009 Jon McPhalen
''               -- portions inspired by code by Mike Green
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 28 JUL 2009
''   Updated.... 29 JUL 2009
''
'' =================================================================================================

{{

   Low/Mid-level I2C routines.
 
   Assumes that SCL and SDA are fixed in a given application and can be set on
   initialization of the object.  Code does not drive SCL or SDA pins high; pull-ups
   are required on both pins.
   
}}   


con

  #0, ACK, NAK


var

  long  scl                                                     ' clock pin of i2c buss
  long  sda                                                     ' data pin of i2c buss
           

pub init(sclpin, sdapin) | okay

'' Define I2C SCL (clock) and SDA (data) pins

  scl := sclpin 
  sda := sdapin


pub start

'' Create I2C start sequence
'' -- will wait if I2C buss SDA pin is held low

  dira[scl] := 0                                                ' float SCL
  repeat while (ina[scl] == 0)                                  ' allow "clock stretching"

  dira[sda] := 0                                                ' SDA high (float to p/u)
  outa[sda] := 0
  dira[sda] := 1                                                ' SDA low 


pub write(i2cbyte) | ackbit

'' Write byte to I2C buss

  outa[scl] := 0
  dira[scl] := 1                                                ' SCL low

  i2cbyte <<= (32-8)                                            ' move msb (bit7) to bit31
  repeat 8                                                      ' output eight bits
    dira[sda] := ((i2cbyte <-= 1) ^ 1)                          ' send msb first
    dira[scl] := 0                                              ' SCL high (float to p/u)
    dira[scl] := 1                                              ' SCL low

  dira[sda] := 0                                                ' relase to read ack bit
  dira[scl] := 0                                                ' SCL high (float) 
  ackbit := ina[sda]                                            ' read ack bit
  dira[scl] := 1                                                ' SCL low

  return ackbit


pub read(ackbit) | i2cbyte

'' Read byte from I2C buss

  outa[scl] := 0                                                ' prep to write low
  dira[sda] := 0                                                ' make input for read

  repeat 8
    dira[scl] := 0                                              ' SCL high (float to p/u)
    i2cbyte := (i2cbyte << 1) | ina[sda]                        ' read the bit
    dira[scl] := 1                                              ' SCL low
                             
  dira[sda] := !ackbit                                          ' output ack bit 
  dira[scl] := 0                                                ' clock it
  dira[scl] := 1

  return i2cbyte & $FF


pub stop

'' Create I2C stop sequence 

  outa[sda] := 0
  dira[sda] := 1                                                ' SDA low
  
  dira[scl] := 0                                                ' float SCL
  repeat while (ina[scl] == 0)                                  ' hold for clock stretch
  
  dira[sda] := 0                                                ' float SDA


pub wait(id) | ackbit

'' Waits for I2C device to be ready for new command

  repeat
    start
    ackbit := write(id & $FE)
  until ackbit == ACK


pub putbyte(id, addr, value) | ackbit

'' Write byte (to device w/8-bit addressing)
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  wait(id)
  write(addr)                                                   ' write address
  ackbit := write(value)                                        ' write one byte
  stop

  return ackbit


pub getbyte(id, addr) | value

'' Read byte (from device w/8-bit addressing)
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  wait(id)
  write(addr)                                                   ' write address
  start                                                         ' restart for read
  write(id | $01)                                               ' read  
  value := read(NAK)                                            '  one byte
  stop

  return value & $0000_00FF
  
  
pub putbytex(id, addr, value) | ackbit

'' Write byte (to device w/16-bit addressing)
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  ackbit := write(value)                                        ' write one byte
  stop

  return ackbit


pub getbytex(id, addr) | value

'' Read byte (from device w/16-bit addressing)
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  start
  write(id | $01)                                               ' read
  value := read(NAK)                                            '  one byte
  stop

  return value & $FF


pub putwordx(id, addr, value) | ackbit

'' Write word (to device w/16-bit addressing) -- storage is Little Endian (lsb, msb)
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
''
'' Be mindful of address in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  ackbit := write(value)                                        ' write lsb
  ackbit |= write(value >> 8)                                   ' write msb
  stop

  return ackbit


pub getwordx(id, addr) | value

'' Read word (from device w/16-bit addressing) - expects Little Endian
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
''
'' Be mindful of address in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  start
  write(id | $01)
  value := read(ACK)                                            ' read lsb
  value |= (read(NAK) << 8)                                     ' read msb  
  stop
 
  return value & $FFFF
 

pub putlongx(id, addr, value) | ackbit

'' Write long (to device w/16-bit addressing) -- storage is Little Endian
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
''
'' Be mindful of address in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  ackbit := write(value)                                        ' lsb (byte0)
  ackbit |= write(value >>  8)
  ackbit |= write(value >> 16)
  ackbit |= write(value >> 24)                                  ' msb (byte3)
  stop

  return ackbit


pub getlongx(id, addr) | value

'' Read long (from device w/16-bit addressing)
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
''
'' Be mindful of address in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address 
  write(addr)                                                   ' lsb of address
  start                                                         ' restart for read
  write(id | $01)                                               ' device read  
  value := read(ACK)                                            ' read lsb (byte0)
  value |= read(ACK) <<  8
  value |= read(ACK) << 16
  value |= read(NAK) << 24                                      ' read msb (byte3)  
  stop

  return value
  

pub putpagex(id, addr, size, src) | ackbit

'' Write block of bytes
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
'' -- size : bytes to write
'' --  src : pointer to the source values
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address
  write(addr)                                                   ' lsb of address
  ackbit := ACK                                                 ' assume okay
  repeat size
    ackbit |= write(byte[src++])                                ' write a byte 
  stop

  return ackbit


pub getpagex(id, addr, size, dest)

'' Read block of bytes
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
'' -- size : bytes to read
'' -- dest : pointer to the destination
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr >> 8)                                              ' msb of address 
  write(addr)                                                   ' lsb of address
  start                                                         ' restart for read
  write(id | $01)                                               ' device read
  repeat size-1
    byte[dest++] := read(ACK)
  byte[dest] := read(NAK)                                       ' last byte gets NAK
  stop


dat

{{

  Copyright (c) 2009 Jon McPhalen (aka Jon Williams)

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}